// Note: The standard SAX library that comes with the JDK
// has a limit of 64000 entity expansions - I don't know why.
// Use the Apache Xerces library instead.
// I used the following command to execute this program:
// java -cp .:xerces-2_10_0/xml-apis.jar:xerces-2_10_0/xercesImpl.jar ToFacts dblp.xml

// This was my first SAX application.
// Some alternatives I tried are commented out.
// Since this is a "one time" program, I did not clean it up.

//import java.io.FileReader;
//import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
//import java.io.PrintStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
//import java.io.PrintWriter;
import java.io.IOException;

//import javax.xml.parsers.FactoryConfigurationError;
//import javax.xml.parsers.ParserConfigurationException;
//import javax.xml.parsers.SAXParser;
//import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.XMLReader;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.helpers.DefaultHandler;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class ToFacts extends DefaultHandler
{

	//=====================================================================
	// Attributes:
	//=====================================================================

	private static OutputStreamWriter out;
	static private int entries;
	static private int facts;

	private String key;
	private boolean factOpen;


	//=====================================================================
	// Main Method:
	//=====================================================================

	public static void main(String args[]) {

		// Check that there is at least one input file:
		if(args.length < 1) {
			System.err.println("This program must be called " +
				"with a DBLP XML input file as " +
				"command line argument.");
				System.exit(1);
		}

		// Create XML Reader:
		XMLReader xr = null;
		try {
			String parserName =
				"org.apache.xerces.parsers.SAXParser";
			xr = XMLReaderFactory.createXMLReader(parserName);
			ToFacts handler = new ToFacts();
			xr.setContentHandler(handler);
			xr.setErrorHandler(handler);
		} catch(Exception e) {
			System.err.println("Exception creating XMLReader: " +
						e);
			e.printStackTrace(System.err);
			System.exit(2);
		}

		// Create SAX Parser:
		//SAXParser xr;
		//try {
		//	SAXParserFactory factory = SAXParserFactory.
		//					newInstance();
		//	xr = factory.newSAXParser();
		//} catch(FactoryConfigurationError e) {
		//	System.err.println("Unable to get a " +
		//				"document builder factory: " +
		//				e);
		//	e.printStackTrace(System.err);
		//	System.exit(2);
		//	return;
		//} catch(ParserConfigurationError e) {
		//	System.err.println("Parser was unable " +
		//				"to be configured: " + e);
		//	e.printStackTrace(System.err);
		//	System.exit(2);
		//	return;
		//} catch(Exception e) {
		//	System.err.println("Exception creating SAXParser: " +
		//				e);
		//	e.printStackTrace(System.err);
		//	System.exit(2);
		//	return;
		//}

		// Create output file:
		try {
			FileOutputStream fsOut =
				new FileOutputStream("out.data");
			out = new OutputStreamWriter(fsOut, "ISO-8859-1");
		} catch(IOException e) {
			System.err.println("Exception opening output file: " +
						e);
			System.exit(3);
		}

		// Initialize number of bibliographic entries processed:
		entries = 0;

		// Initialize number of facts produced:
		facts = 0;

		// Parse each file provided on the command line.
		for (int i = 0; i < args.length; i++) {
			InputStreamReader in = null;
			try {
				FileInputStream is =
					new FileInputStream(args[i]);
				in = new InputStreamReader(is, "ISO-8859-1");
			} catch(IOException e) {
				System.err.println("Error opening input file "+
					"'" + args[i] + "': " + e);
				System.exit(4);
			}
			try {
				xr.parse(new InputSource(in));
				//xr.parse(new InputSource(in), new ToFacts());
			} catch(Exception e) {
				System.err.println("Error parsing " +
					"'" + args[i] + "': " + e);
				System.exit(5);
			}
		}

		// Close Output File:
		try {
			out.close();
		} catch(IOException e) {
			System.err.println("Exception closing output file: " +
						e);
			System.exit(3);
		}

		// Show statistics:
		double facts_per_entry = ((double)facts)/entries;
		System.out.println("Number of entries processed: " + entries);
		System.out.println("Number of facts written:     " + facts);
		System.out.format( "Facts per Entry:              %5.1f\n",
				new Object[]{new Double(facts_per_entry)});
	}


	//=====================================================================
	// Constructor:
	//=====================================================================

	public ToFacts() {
		super();
	}


	//=====================================================================
	// Auxiliary Functions:
	//=====================================================================

	private boolean isBibType(String name) {
		return	name.equals("article") ||
			name.equals("inproceedings") ||
			name.equals("proceedings") ||
			name.equals("book") ||
			name.equals("incollection") ||
			name.equals("phdthesis") ||
			name.equals("mastersthesis") ||
			name.equals("www") ||
			name.equals("person") ||
			name.equals("data");
	}


	private boolean isBibField(String name) {
		return	name.equals("author") ||
			name.equals("title") ||
			name.equals("year") ||
			name.equals("month") ||
			name.equals("journal") ||
			name.equals("volume") ||
			name.equals("number") ||
			name.equals("pages") ||
			name.equals("editor") ||
			name.equals("booktitle") ||
			name.equals("series") ||
			name.equals("chapter") ||
			name.equals("publisher") ||
			name.equals("address") ||
			name.equals("isbn") ||
			name.equals("url") ||
			name.equals("ee") ||
			name.equals("cdrom") ||
			name.equals("note") ||
			name.equals("cite") ||
			name.equals("crossref") ||
			name.equals("publnr") ||
			name.equals("school");
	}

	private boolean isTextMarkup(String name) {
		return	name.equals("i") ||
			name.equals("sub") ||
			name.equals("sup") ||
			name.equals("tt") ||
			name.equals("ref");
	}

	private void writeFact(String fact) {
		facts++;
		try {
			ToFacts.out.write(fact + "\n");
		} catch(IOException e) {
			System.err.println("Error writing output: "+e);
		}
	}

	private void writeFactStart(String fact) {
		facts++;
		try {
			ToFacts.out.write(fact);
		} catch(IOException e) {
			System.err.println("Error writing output: "+e);
		}
	}

	private void writeFactText(String str) {
		try {
			ToFacts.out.write(str);
		} catch(IOException e) {
			System.err.println("Error writing output: "+e);
		}
	}

	private void writeFactChar(int ch) {
		try {
			ToFacts.out.write(ch);
		} catch(IOException e) {
			System.err.println("Error writing output: "+e);
		}
	}

	private void writeFactEnd(String str) {
		try {
			ToFacts.out.write(str + "\n");
		} catch(IOException e) {
			System.err.println("Error writing output: "+e);
		}
	}

	//=====================================================================
	// Event Handlers:
	//=====================================================================


	public void startDocument() {
		//System.out.println("Start document");
	}


	public void endDocument() {
		//System.out.println("End document");
	}


	public void startElement(String uri, String name,
			      String qName, Attributes attrs) {

		//if ("".equals (uri))
		//	System.out.println("Start element: " + qName);
		//else
		//	System.out.println("Start element: {" + uri + "}" +
		//				name);
		//int attrCount = attrs.getLength();
		//if(attrCount>0) {
		//	System.out.println("Attributes:"); 
		//	for(int i = 0 ; i<attrCount ; i++) {
		//		System.out.println("  Name : " +
		//					attrs.getQName(i)); 
		//		System.out.println("  Type : " +
		//					attrs.getType(i)); 
		//		System.out.println("  Value: " +
		//					attrs.getValue(i)); 
		//	}
		//}

		if(name == null || name.equals(""))
			name = qName;

		if(name.equals("dblp")) {
			// do nothing
		}
		else if(isBibType(name)) {
			this.key = attrs.getValue("key");
			if(this.key == null) {
				System.err.println("Element of type '" + name +
						"' without key attribute " +
						"(ignored)");
				return;
			}
			String fact = "att('" + key + "', type, " +
						"'" + name + "').";
			writeFact(fact);
			int attrCount = attrs.getLength();
			for(int i = 0 ; i < attrCount ; i++) {
				String attr = attrs.getLocalName(i);
				if(attr != null && !attr.equals("key")) {
					fact = "att('" + key +
						"', " + attr + ", '" +
						attrs.getValue(i) + "').";
					writeFact(fact);
				}
			}
		}
		else if(isBibField(name)) {
			if(this.key != null) {
				writeFactStart(
					"att('" + key + "', " + name + ", '");
				this.factOpen = true;
			}
			else {
				this.factOpen = false;
			}
		}
		else if(isTextMarkup(name)) {
			if(this.factOpen) {
				if(name.equals("ref")) {
					String href = attrs.getValue("href");
					if(href == null) {
						System.err.println(
							"ref without href");
						return;
					}
					writeFactText("<ref href=\"" +
							href + "\">");
				}
				else {
					writeFactText("<" + name + ">");
				}
			}
			else {
				System.err.println("Text markup '" + name +
					"' outside text content.");
			}
		}
		else {
			System.err.println("Unknown element: '" + name + "'");
		}
	}


	public void endElement(String uri, String name, String qName) {
		//if ("".equals (uri))
		//	System.out.println("End element: " + qName);
		//else
		//	System.out.println("End element:   {" + uri + "}" +
		//				name);

		if(name == null || name.equals(""))
			name = qName;

		if(isTextMarkup(name)) {
			if(this.factOpen) {
				writeFactText("</" + name + ">");
			}
		}
		else if(this.factOpen) {
			writeFactEnd("').");
			this.factOpen = false;
		}

		if(isBibType(name)) {
			this.key = null;
			entries++;
			if(entries % 100000 == 0)
				System.out.println(entries +
						" entries processed");
		}
	}


	public void characters(char ch[], int start, int length) {
		//System.out.print("Characters:    \"");
		//for (int i = start; i < start + length; i++) {
		//	switch (ch[i]) {
		//	case '\\':
		//		System.out.print("\\\\");
		//		break;
		//	case '"':
		//		System.out.print("\\\"");
		//		break;
		//	case '\n':
		//		System.out.print("\\n");
		//		break;
		//	case '\r':
		//		System.out.print("\\r");
		//		break;
		//	case '\t':
		//		System.out.print("\\t");
		//		break;
		//	default:
		//		System.out.print(ch[i]);
		//		break;
		//	}
		//}
		//System.out.print("\"\n");

		for (int i = start; i < start + length; i++) {
			switch (ch[i]) {
			case '\\':
				writeFactText("\\\\");
				break;
			case '"':
				writeFactText("\\\"");
				break;
			case '\'':
				writeFactText("\\\'");
				break;
			case '\n':
				writeFactText("\\n");
				break;
			case '\r':
				writeFactText("\\r");
				break;
			case '\t':
				writeFactText("\\t");
				break;
			default:
				writeFactChar(ch[i]);
				break;
			}
		}
	}

	public void warning(SAXParseException spe) {
		System.err.println("Warning at line "+spe.getLineNumber());
		System.err.println(spe.getMessage());
	}

	public void fatalError(SAXParseException spe) throws SAXException {
		System.err.println("Fatal error at line "+spe.getLineNumber());
		System.err.println(spe.getMessage());
		throw spe;
	}

}
